#include <inttypes.h>
#include <string.h>
#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/interrupt.h>

// Inline assembly: The nop = do nothing for one clock cycle.
#define nop()  __asm__ __volatile__("nop")

#ifndef __cplusplus
typedef unsigned char bool;
enum
{
	false=0, true=1
};
#endif

// Define some useful types:
typedef signed char int8;
typedef uint8_t uint8, byte;
typedef uint16_t uint16, word;
typedef uint32_t uint32, dword;

void Wait(uint32_t n)
{
	while (n--)
		nop();
}

#define F_CPU 16000000

void uart_init(word  baud)
{
	word ubrr=(F_CPU / 16 / baud - 1);

	UBRRH = (uint8_t)(ubrr >> 8);
	UBRRL = (uint8_t)ubrr;
 

	UCSRA = 0;//(1 << U2X);
	UCSRB = (1 << RXEN) | (1 << TXEN);
 
	UCSRC = (1 << URSEL) | (1<<USBS) | (3 << UCSZ0);

	sei();

}

enum
{
	Executing=1,
	Paused=2,
	PauseRequested=4,
};

word g_num=0;
struct
{
	dword x, y, z;
	byte px, py, pz;
	byte status;
	int32_t vx, vy, vz;
	int32_t ax, ay, az;
	int32_t aax, aay, aaz;
	dword finalX, finalY, finalZ;
	dword ticks;
} machine;

dword v=0;


typedef struct
{
	dword id;
	int32_t vx, vy, vz;
	int32_t ax, ay, az;
	int32_t aax, aay, aaz;
	dword x, y, z;
	dword ticks;
	word checksum;
} Packet;

bool ValidatePacket(const Packet *packet)
{
	const byte *p=(const byte *)packet;
	word chk=0;
	for (uint8_t i=0; i<sizeof(Packet)-2; i++)
	{
		chk^=((chk+0x5ce6)<<1)+p[i];
	}
	return chk==packet->checksum;
}

enum
{
	PacketSize=sizeof(Packet),
	MinPackets=16,
	MaxPackets=24
};


Packet Packets[MaxPackets];

static inline void SetOutputs()
{
	static byte d=0;
	byte s=0;

	byte qx=machine.x>>24;
	byte qy=machine.y>>24;
	byte qz=machine.z>>24;

	int8_t dx=qx-machine.px;
	int8_t dy=qy-machine.py;
	int8_t dz=qz-machine.pz;


	if (dx<0) { d|=2; s|=1; }
	else if (dx) { d&=~2; s|=1; }
	if (dy<0) { d|=8; s|=4; }
	else if (dy) { d&=~8; s|=4; }
	if (dz<0) { d|=32; s|=16; }
	else if (dz) { d&=~32; s|=16; }


	PORTA=d;
	machine.px=qx;
	machine.py=qy;
	machine.pz=qz;
	PORTA=d|s;


}


#define USART_RX_BUFFER_SIZE 128     /* 2,4,8,16,32,64,128 or 256 bytes */
#define USART_TX_BUFFER_SIZE 128     /* 2,4,8,16,32,64,128 or 256 bytes */
#define USART_RX_BUFFER_MASK ( USART_RX_BUFFER_SIZE - 1 )
#define USART_TX_BUFFER_MASK ( USART_TX_BUFFER_SIZE - 1 )

static byte USART_TxBuf[USART_TX_BUFFER_SIZE];
static volatile byte USART_TxHead=0;
static volatile byte USART_TxTail=0;

ISR(USART_UDRE_vect)
{
	unsigned char tmptail;

	// Check if all data is transmitted
	if ( USART_TxHead != USART_TxTail )
	{
		// Calculate buffer index 
		tmptail = ( USART_TxTail + 1 ) & USART_TX_BUFFER_MASK;
		USART_TxTail = tmptail;      // Store new index
	
		UDR = USART_TxBuf[tmptail];  // Start transmition
	}
	else
	{
		UCSRB &= ~(1<<UDRIE);         // Disable UDRE interrupt
	}
}

void TransmitByte(byte data)
{
	byte tmphead;
	tmphead = ( USART_TxHead + 1 ) & USART_TX_BUFFER_MASK;
	while (tmphead == USART_TxTail)
		;

	USART_TxBuf[tmphead] = data;
	USART_TxHead = tmphead;

	UCSRB |= (1<<UDRIE);
}

void Transmit(void *buffer, byte n)
{
	for (byte i=0; i<n; i++)
	{
		TransmitByte(((byte*)buffer)[i]);
	}
}

typedef enum
{
	R_Invalid=0,
	R_More=1,
	R_QueueFull=2,
	R_Pause=3,
	R_Finished=100,
	R_Error=128,
} Response;

inline void Ack(dword id)
{
	TransmitByte(R_More);
	Transmit(&id, 4);
}

inline void Nack()
{
	TransmitByte(R_Error);
}

inline void QueueFull()
{
	TransmitByte(R_QueueFull);
}

typedef enum
{
	None=0,
	Curve=0x5,
	GetStatus=0xa,
	Commit=0x80,
	Pause=0x85,
	Resume=0x8a,
	Reset=0xff,
} PacketType;

struct
{
	byte w;
	PacketType state;
	byte full;
	dword LastId;

	byte packets, write, read;

	bool drop;

} Receiver;

void ResetMachine()
{
	memset(&machine, 0, sizeof(machine));
	memset(&Receiver, 0, sizeof(Receiver));
	
}

void ProcessByte(byte value)
{
	switch (Receiver.state)
	{
	case None:
		switch ((PacketType)value)
		{
		case Commit:
			machine.status|=Executing;
			break;

		case Pause:
			machine.status|=PauseRequested;
			break;

		case Resume:
			machine.status&=~(Paused|PauseRequested);
			break;

		case Reset:
			ResetMachine();
			break;

		case Curve:
			Receiver.drop=false;
			Receiver.state=(PacketType)value;
			break;

		default:
			Receiver.state=None;
			break;
		}
		break;

	case Curve:
		if (Receiver.full)
			Receiver.drop=true;
		if (Receiver.drop)
			Receiver.w++; // drop the transmission
		else
			((byte*)&(Packets[Receiver.write]))[Receiver.w++]=value;

		if (Receiver.w>=PacketSize)
		{
			Receiver.state=None;

			Receiver.w=0;
			if (Receiver.drop)
			{
				QueueFull();
				break;
			}

			if (!ValidatePacket(&Packets[Receiver.write]))
			{

				Nack();
				break;
			}

			if (Packets[Receiver.write].id==Receiver.LastId)
			{
				// same Id - probably a resend; ignore
				Ack(Receiver.LastId);
				break;
			}

			if (Receiver.packets<MaxPackets)
			{
				Receiver.LastId=Packets[Receiver.write].id;
				if (Receiver.packets<MaxPackets-1)
					Ack(Receiver.LastId);
				else
				{
					Receiver.full=1;
					QueueFull();
	            }
				if (++Receiver.write==MaxPackets)
					Receiver.write=0;
				Receiver.packets++;
				PORTB|=1;
			}
			else
			{
				Receiver.full=1;
				QueueFull();
			}
		}
		break;

	default:
		Receiver.state=None;
		break;
	}
}


int main(void)
{
	DDRA = 0xff; // Set Port A as output
	DDRB = 0xff; // Set Port B as output
	PORTA = 0;
	PORTB = 0xff;
	Wait(1000000);

	uart_init(50000);


	for (;;)
	{

		//for (byte b=0; (UCSRA & (1<<RXC)) && b<8; b++)
		if (UCSRA & (1<<RXC))
		{
			ProcessByte(UDR);
		}

		if (!machine.ticks)
		{

			machine.x=machine.finalX;
			machine.y=machine.finalY;
			machine.z=machine.finalZ;


	
			if (!Receiver.packets)
			{
				SetOutputs();
				PORTB&=0xfe;
				machine.status&=~Executing;
				PORTB|=2;
				continue; // underflow
			}


			if (machine.status&PauseRequested)
				machine.status=(machine.status&~PauseRequested)|Paused;
			if (machine.status&Paused)
				continue;
			if (!(machine.status&Executing))
			{
				PORTB|=2;
				if (Receiver.packets>=MinPackets)
					machine.status|=Executing;
	            else
					continue;
			}
			else
				PORTB&=0xfd;


			
			Receiver.packets--;
			if (Receiver.full)
			{
				Ack(Receiver.LastId);
				Receiver.full=0;
			}


			machine.vx=Packets[Receiver.read].vx;
			machine.vy=Packets[Receiver.read].vy;
			machine.vz=Packets[Receiver.read].vz;
			machine.ax=Packets[Receiver.read].ax;
			machine.ay=Packets[Receiver.read].ay;
			machine.az=Packets[Receiver.read].az;
			machine.aax=Packets[Receiver.read].aax;
			machine.aay=Packets[Receiver.read].aay;
			machine.aaz=Packets[Receiver.read].aaz;
			machine.finalX=Packets[Receiver.read].x;
			machine.finalY=Packets[Receiver.read].y;
			machine.finalZ=Packets[Receiver.read].z;
			machine.ticks=Packets[Receiver.read].ticks;

			if (++Receiver.read>=MaxPackets)
				Receiver.read=0;
		}
		SetOutputs();

		if (--machine.ticks)
		{
			machine.x+=machine.vx>>8;
			machine.y+=machine.vy>>8;
			machine.z+=machine.vz>>8;
			machine.vx+=machine.ax>>8;
			machine.vy+=machine.ay>>8;
			machine.vz+=machine.az>>8;
			machine.ax+=machine.aax;
			machine.ay+=machine.aay;
			machine.az+=machine.aaz;
		}


	}

	return 0;

}
