#pragma once
#ifndef _BEZIER_H
#define _BEIZER_H

#include "Vector.h"

template <class Vector, class scalar>
class CubicBezier
{
public:

    inline CubicBezier() {}
    inline CubicBezier(const Vector &_A, const Vector &_B, const Vector &_C, const Vector &_D)
        : A(_A), B(_B), C(_C), D(_D)
    {}

    Vector A, B, C, D;

    mutable Vector a, b, c, d;

    inline void CalculateCoefficients() const
    {
        a=3.0*(B-C)-A+D;
        b=3.0*(A+C-B-B);
        c=3.0*(B-A);
        d=A;
    }

    inline Vector operator()(scalar t) const
    {
        scalar t2=t*t;
        scalar t3=t2*t;
        return a*t3+b*t2+c*t+d;
    }

    inline scalar Length() const
    {
        if (!c.LengthSquare() && !d.LengthSquare())
            return (D-A).Length();

        int steps=1000;

        Vector prev=A;
        scalar l=0;
        for (int i=1; i<=steps; i++)
        {
            Vector V=(*this)((double)i/steps);
            l+=(V-prev).Length();
            prev=V;
        }
        return l;
    }

    inline void Split(scalar t, CubicBezier &first, CubicBezier &second) const
    {
        Vector P=Lerp(A, B, t);
        Vector Q=Lerp(B, C, t);
        Vector R=Lerp(C, D, t);
        Vector U=Lerp(P, Q, t);
        Vector V=Lerp(Q, R, t);
        Vector X=Lerp(U, V, t);

        // use own data first, so the procedure works when one of first/second is *this
        first.A=A;
        second.D=D;

        first.B=P;
        first.C=U;
        first.D=X;

        second.A=X;
        second.B=V;
        second.C=R;
    }

    inline CubicBezier operator*(double m) const
    {
        CubicBezier r;
        r.A=A*m;
        r.B=B*m;
        r.C=C*m;
        r.D=D*m;
        r.a=a*m;
        r.b=b*m;
        r.c=c*m;
        r.d=d*m;
        return r;
    }

    static inline Vector Lerp(const Vector &A, const Vector &B, scalar t)
    {
        return A+(B-A)*t;
    }

    inline void Reverse()
    {
        V3f tmp=A;
        A=D; D=tmp;
        tmp=B;
        B=C; C=tmp;
        CalculateCoefficients();
    }

};

typedef CubicBezier<V3f, double> Bezier3D;


double CurveError(const Bezier3D &a, const Bezier3D &b, int steps=1000)
{
    V3f *bpoints=new V3f[steps+1];
    for (int i=0; i<=steps; i++)
    {
        bpoints[i]=b((double)i/steps);
    }

    int prevj=0;

    double e=0;
    for (int i=0; i<=steps; i++)
    {
        V3f X=a((double)i/steps);
        double d=1e+30;
        int minj=prevj;
        int startj=prevj-steps/10;
        int endj=prevj+steps/10;
        if (startj<0) startj=0;
        if (endj>steps) endj=steps;
        for (int j=startj; j<=endj; j++)
        {
            double dd=(X-bpoints[j]).LengthSquare();
            if (dd<d)
            {
                d=dd;
                minj=j;
            }
        }
        prevj=minj;
        if (d>e)
            e=d;

    }

    delete[] bpoints;
    return sqrt(e);
}



double BezierIntersectionXY(const Bezier3D &A, const Bezier3D &B, double &ta, double &tb)
{
    const int n=300;
    V3f _a[n+1], _b[n+1];
    A.CalculateCoefficients();
    B.CalculateCoefficients();
    for (int i=0; i<=n; i++)
    {
        _a[i]=A((double)i/n);
        _b[i]=B((double)i/n);
    }

    double min=1e+30;
    int mini=0, minj=0;

    for (int i=0; i<=n; i++)
    {
        for (int j=n; j>=0; j--)
        {
            double dx=_a[i].X-_b[j].X;
            double dy=_a[i].Y-_b[j].Y;
            double d=dx*dx+dy*dy;
            if (d<min)
            {
                mini=i;
                minj=j;
                min=d;
            }
        }
    }

    double ta0=(double)(mini-1)/n;
    double ta1=(double)(mini+1)/n;
    double tb0=(double)(minj-1)/n;
    double tb1=(double)(minj+1)/n;

    if (ta0<0) ta0=0;
    if (ta1>1) ta1=1;
    if (tb0<0) tb0=0;
    if (tb1>1) tb1=1;
    if (ta0==ta1 && tb0==tb1)
        return min;

    for (int i=0; i<=n; i++)
    {
        _a[i]=A(ta0+(ta1-ta0)*i/n);
        _b[i]=B(tb0+(tb1-tb0)*i/n);
    }

    for (int i=0; i<=n; i++)
    {
        for (int j=n; j>=0; j--)
        {
            double dx=_a[i].X-_b[j].X;
            double dy=_a[i].Y-_b[j].Y;
            double d=dx*dx+dy*dy;
            if (d<min)
            {
                mini=i;
                minj=j;
                min=d;
            }
        }
    }

    double pta0=ta0;
    double ptb0=tb0;
    ta0=pta0+(ta1-pta0)*(mini-1)/n;
    ta1=pta0+(ta1-pta0)*(mini+1)/n;
    tb0=ptb0+(tb1-ptb0)*(minj-1)/n;
    tb1=ptb0+(tb1-ptb0)*(minj+1)/n;

    if (ta0<0) ta0=0;
    if (ta1>1) ta1=1;
    if (tb0<0) tb0=0;
    if (tb1>1) tb1=1;
    if (ta0==ta1 && tb0==tb1)
        return min;

    for (int i=0; i<=n; i++)
    {
        _a[i]=A(ta0+(ta1-ta0)*i/n);
        _b[i]=B(tb0+(tb1-tb0)*i/n);
    }

    for (int i=0; i<=n; i++)
    {
        for (int j=n; j>=0; j--)
        {
            double dx=_a[i].X-_b[j].X;
            double dy=_a[i].Y-_b[j].Y;
            double d=dx*dx+dy*dy;
            if (d<min)
            {
                mini=i;
                minj=j;
                min=d;
            }
        }
    }

    pta0=ta0;
    ptb0=tb0;
    ta0=pta0+(ta1-pta0)*(mini-1)/n;
    ta1=pta0+(ta1-pta0)*(mini+1)/n;
    tb0=ptb0+(tb1-ptb0)*(minj-1)/n;
    tb1=ptb0+(tb1-ptb0)*(minj+1)/n;

    if (ta0<0) ta0=0;
    if (ta1>1) ta1=1;
    if (tb0<0) tb0=0;
    if (tb1>1) tb1=1;
    if (ta0==ta1 && tb0==tb1)
        return min;

    for (int i=0; i<=n; i++)
    {
        _a[i]=A(ta0+(ta1-ta0)*i/n);
        _b[i]=B(tb0+(tb1-tb0)*i/n);
    }

    for (int i=0; i<=n; i++)
    {
        for (int j=n; j>=0; j--)
        {
            double dx=_a[i].X-_b[j].X;
            double dy=_a[i].Y-_b[j].Y;
            double d=dx*dx+dy*dy;
            if (d<min)
            {
                mini=i;
                minj=j;
                min=d;
            }
        }
    }

    ta=ta0+(ta1-ta0)*mini/n;
    tb=tb0+(tb1-tb0)*minj/n;

    return min;
}


double Nearest(const Bezier3D &B, const V3f &P)
{
    const int n=300;
    B.CalculateCoefficients();

    double min=1e+30;
    int mini=-1, minj=-1;

    for (int i=0; i<=n; i++)
    {
        V3f pt=B((double)i/n);
        double dx=pt.X-P.X;
        double dy=pt.Y-P.Y;
        double d=dx*dx+dy*dy;
        if (d<min)
        {
            mini=i;
            min=d;
        }
    }
    if (mini<0)
        return -1;

    double ta0=(double)(mini-1)/n;
    double ta1=(double)(mini+1)/n;
    if (ta0<0) { if (ta1<=0) return 0; } ta0=0;
    if (ta1>1) { if (ta0>=1) return 1; } ta1=1;

    for (int i=0; i<=n; i++)
    {
        V3f pt=B(ta0+(ta1-ta0)*i/n);
        double dx=pt.X-P.X;
        double dy=pt.Y-P.Y;
        double d=dx*dx+dy*dy;
        if (d<min)
        {
            mini=i;
            min=d;
        }
    }

    double pta0=ta0;
    ta0=pta0+(ta1-pta0)*(mini-1)/n;
    ta1=pta0+(ta1-pta0)*(mini+1)/n;
    if (ta0<0) { if (ta1<=0) return 0; } ta0=0;
    if (ta1>1) { if (ta0>=1) return 1; } ta1=1;

    for (int i=0; i<=n; i++)
    {
        V3f pt=B(ta0+(ta1-ta0)*i/n);
        double dx=pt.X-P.X;
        double dy=pt.Y-P.Y;
        double d=dx*dx+dy*dy;
        if (d<min)
        {
            mini=i;
            min=d;
        }
    }

    ta0=ta0+(ta1-ta0)*mini/n;
    return ta0<0 ? 0 : ta0>1 ? 1 : ta0;
}



int IsContinuous(const Bezier3D &first, const Bezier3D &second, double dist_tolerance=0, double angle_tolerance=5 /*degrees*/)
{
    if (first.D!=second.A)
    {
        if ((first.D-second.A).LengthSquare()>dist_tolerance*dist_tolerance)
            return 0;
    }

    V3f cd1=(first.D-first.C);
    if (!cd1.LengthSquare())
        return 1; // full stop at the end of first curve

    V3f ab2=(second.B-second.A);
    if (!ab2.LengthSquare())
        return 1; // full stop at the start of second curve

    double cosine=(cd1*ab2)/(cd1.Length()*ab2.Length());
    if (cosine<cos(angle_tolerance*M_PI/180))
        return 1;

    return 2;
}


#endif

