#pragma once
#ifndef _CNC_H
#define _CNC_H

#include <windows.h>
#undef max
#undef min
#include <math.h>
#include "Bezier.h"
#include "Vector.h"
#include <vector>
#include <iostream>
#include <ftd2xx.h>

typedef unsigned int dword;
typedef unsigned short word;

enum PacketType
{
	Packet_None=0,
	Packet_Curve=0x5,
	Packet_GetStatus=0xa,
	Packet_Commit=0x80,
	Packet_Pause=0x85,
	Packet_Resume=0x8a,
	Packet_Reset=0xff,
};

#pragma pack(push, 1)

struct Packet
{
    dword id;
	int vx, vy, vz;
	int ax, ay, az;
	int aax, aay, aaz;
	int x, y, z;
	int ticks;
    word checksum;
};

#pragma pack(pop)

void SetChecksum(Packet *packet)
{
	const unsigned char *p=(const unsigned char *)packet;
	word chk=0;
	for (int i=0; i<sizeof(Packet)-2; i++)
	{
		chk^=((chk+0x5ce6)<<1)+p[i];
	}
	packet->checksum=chk;
}



enum ResponseStatus
{
	R_Invalid=0,
	R_More=1,
	R_QueueFull=2,
	R_Pause=3,
	R_Finished=100,
	R_Error=128,
    R_Timeout=254,
    R_CommError=255
};

struct MachineResponse
{
    ResponseStatus Status;
    int PacketId;
};

#pragma warning(disable:4244)

template <class T>
T max(const T &a, const T &b, const T &c)
{
    return b>a ? c>b ? c : b : c>a ? c : a;
}
template <class T>
T min(const T &a, const T &b, const T &c)
{
    return b<a ? c<b ? c : b : c<a ? c : a;
}

class MotionExecutor
{
public:
    MotionExecutor()
    {
        Div=V3f(-800, -800, 800);
        MachineFrequency=40000;
        PacketId=0;
        h=0;
        verbose=false;
    }

    ~MotionExecutor()
    {
        Close();
    }

    bool Open(int baud)
    {
        Close();

        char buf[1024];
        FT_STATUS status = FT_ListDevices(0, buf, FT_LIST_BY_INDEX | FT_OPEN_BY_SERIAL_NUMBER);
	    if (status!=FT_OK) //Get first device serial number
	    {
            return false;
	    }

        status=FT_OpenEx(buf, FT_OPEN_BY_SERIAL_NUMBER, &h);

        FT_ResetDevice(h);
        Sleep(100);
        status=FT_SetBaudRate(h, 50000);
        status=FT_SetDataCharacteristics(h, 8, 2, 0);

        //status=FT_SetTimeouts(h, -1, -1);

        return true;
    }

    void Close()
    {
        FT_Close(h);
        h=0;
    }

    V3f Div;

    double MachineFrequency;
    int PacketId;

    bool Wait(double time)
    {
        Packet P={0};
        P.ticks=(int)(time*MachineFrequency);

        return WritePacket(Packet_Curve, P);
    }

    bool WritePacketEx(PacketType type, const Packet &P)
    {
        DWORD w;
        BYTE _type=type;
        FT_STATUS status;
        status=FT_Write(h, &_type, 1, &w);
        if (status!=FT_OK || w!=1)
            return false;
        status=FT_Write(h, (LPVOID)&P, sizeof(P), &w);
        if (status!=FT_OK || w!=sizeof(P))
            return false;
        return true;
    }

    bool WritePacket(PacketType type, Packet &P)
    {
        P.id=++PacketId;
        SetChecksum(&P);

        if (!WritePacketEx(type, P))
            return false;

        int attempts=10;
        for (int i=0;;)
        {
            MachineResponse mr=ReadResponse();

            if (mr.Status==R_QueueFull)
                continue;
            if (mr.Status==R_Error || mr.Status==R_Timeout)
            {
                if (!WritePacketEx(type, P))
                    return false;
                if (i++<attempts)
                    continue;
                else
                    break;
            }
            if (mr.Status==R_More && mr.PacketId!=P.id)
            {
                std::cerr << "Last acknowledged packet is not this one:" << P.id << std::endl;
                std::cerr << "Resending" << std::endl;
                if (!WritePacketEx(type, P))
                    return false;
                continue;
            }
            if (mr.Status!=R_More)
            {
                std::cerr << "Machine error!" << std::endl;
                return false;
            }
            break;
        }
        return true;
    }

    void Commit()
    {
        DWORD w;
        BYTE type=Packet_Commit;
        FT_Write(h, &type, 1, &w);
    }
    void Reset()
    {
        DWORD w;
        BYTE type=Packet_Reset;
        FT_Write(h, &type, 1, &w);
    }

    void Pause()
    {
        DWORD w;
        BYTE type=Packet_Pause;
        FT_Write(h, &type, 1, &w);
    }
    void Resume()
    {
        DWORD w;
        BYTE type=Packet_Resume;
        FT_Write(h, &type, 1, &w);
    }


    MachineResponse ReadResponse()
    {
        MachineResponse mr={R_Invalid};

        BYTE type;
        DWORD w;
        if (FT_Read(h, &type, 1, &w)!=FT_OK)
        {
            mr.Status=R_CommError;
            return mr;
        }
        if (!w)
        {
            mr.Status=R_Timeout;
            return mr;
        }

        switch (type)
        {
        case R_More:
            {
                int rd=0;
                union
                {
                    int id;
                    char buf[4];
                };
                while (rd<4)
                {
                    if (FT_Read(h, buf+rd, 4-rd, &w)!=FT_OK)
                    {
                        mr.Status=R_CommError;
                        return mr;
                    }
                    rd+=w;
                }
                if (verbose)
                    std::cerr << "Send more; last = " << id << std::endl;
                mr.PacketId=id;
            }
            break;
        case R_QueueFull:
            if (verbose)
                std::cerr << "Queue full" << std::endl;
            break;
        default:
            std::cerr << "Resp: " << (int)type << std::endl;
            break;
        }

        mr.Status=(ResponseStatus)type;
        return mr;
    }


    bool ExecuteCurve(const V3f &A, const V3f &B, const V3f &C, const V3f &D, double time)
    {
        Packet P=PrepareCurve(A, B, C, D, time);
        if (P.ticks<=0)
            return true;
        return WritePacket(Packet_Curve, P);
    }


    Packet PrepareCurve(const V3f &A, const V3f &B, const V3f &C, const V3f &D, float time)
    {
        Packet P={0};
        P.ticks=(int)ceil(MachineFrequency*time);
        
        V3f a=ToMachine(3.0*(B-C)-A+D);
        V3f b=ToMachine(3.0*(A+C-B-B));
        V3f c=ToMachine(3.0*(B-A));
        V3f d=ToMachine(A);
        V3f dd=ToMachine(D);

        V3f aaa=6.0*a;
        V3f a0=b+b;
        V3f a1=a0+aaa;
        const V3f &v0=c;
        V3f v1=a*3.0+a0+c;

        double vmul=4294967296.0;
        P.vx=v0.X*vmul/P.ticks;
        P.vy=v0.Y*vmul/P.ticks;
        P.vz=v0.Z*vmul/P.ticks;

        double tt=(long long)P.ticks*P.ticks;
        double amul=(long long)1<<40;
        P.ax=a0.X*amul/tt;
        P.ay=a0.Y*amul/tt;
        P.az=a0.Z*amul/tt;

        double ttt=tt*P.ticks;
        double aamul=(long long)1<<40;
        P.aax=aaa.X*aamul/ttt;
        P.aay=aaa.Y*aamul/ttt;
        P.aaz=aaa.Z*aamul/ttt;

        double posmul=1<<24;

        long long finalX=dd.X*posmul;
        long long finalY=dd.Y*posmul;
        long long finalZ=dd.Z*posmul;


        P.x=finalX;
        P.y=finalY;
        P.z=finalZ;

        CorrectOvershoot(P, v1, d.X*posmul, d.Y*posmul, d.Z*posmul, finalX, finalY, finalZ);
        CheckDoubleStep(P, v1, d.X*posmul, d.Y*posmul, d.Z*posmul, finalX, finalY, finalZ);

        return P;
    }

    void CorrectOvershoot(Packet &P, const V3f &v1, long long startX, long long startY, long long startZ, long long finalX, long long finalY, long long finalZ)
    {
        bool isOvershot=true;

        double tt=(long long)P.ticks*P.ticks;
        for (int tries=0; tries<100 && isOvershot; tries++)
        {
            isOvershot=false;

            long long x, y, z;
            int vx, vy, vz;
            int ax, ay, az;
            int aax, aay, aaz;
            x=startX;
            y=startY;
            z=startZ;
            vx=P.vx;
            vy=P.vy;
            vz=P.vz;
            ax=P.ax;
            ay=P.ay;
            az=P.az;
            aax=P.aax;
            aay=P.aay;
            aaz=P.aaz;
            for (int i=0; i<P.ticks; i++)
            {
			    x+=vx>>8;
			    y+=vy>>8;
			    z+=vz>>8;
			    vx+=ax>>8;
			    vy+=ay>>8;
			    vz+=az>>8;
			    ax+=aax;
			    ay+=aay;
			    az+=aaz;
            }

            if ((x-finalX)>>24)
            {
                P.ax-=((x-finalX)<<17)/tt;
                isOvershot=true;
            }
            if ((y-finalY)>>24)
            {
                P.ay-=((y-finalY)<<17)/tt;
                isOvershot=true;
            }
            if ((z-finalZ)>>24)
            {
                P.az-=((z-finalZ)<<17)/tt;
                isOvershot=true;
            }
        }
    }

    void CheckDoubleStep(Packet &P, const V3f &v1, long long startX, long long startY, long long startZ, long long finalX, long long finalY, long long finalZ)
    {
        long long x, y, z;
        long long px, py, pz;
        int vx, vy, vz;
        int ax, ay, az;
        int aax, aay, aaz;
        x=startX;
        y=startY;
        z=startZ;
        vx=P.vx;
        vy=P.vy;
        vz=P.vz;
        ax=P.ax;
        ay=P.ay;
        az=P.az;
        aax=P.aax;
        aay=P.aay;
        aaz=P.aaz;
        for (int i=0; i<P.ticks; i++)
        {
            px=x>>24;
            py=y>>24;
            pz=z>>24;
		    x+=vx>>8;
		    y+=vy>>8;
		    z+=vz>>8;
		    vx+=ax>>8;
		    vy+=ay>>8;
		    vz+=az>>8;
		    ax+=aax;
		    ay+=aay;
		    az+=aaz;

            if (abs((int)(px-(x>>24)))>1 ||
                abs((int)(py-(y>>24)))>1 ||
                abs((int)(pz-(z>>24)))>1)
            {
                std::cerr << "Steps will be lost" << std::endl;
            }
        }

        px=x>>24;
        py=y>>24;
        pz=z>>24;
        x=finalX;
        y=finalY;
        z=finalZ;

        if (abs((int)(px-(x>>24)))>1 ||
            abs((int)(py-(y>>24)))>1 ||
            abs((int)(pz-(z>>24)))>1)
        {
            std::cerr << "Steps will be lost" << std::endl;
        }


    }

    inline V3f ToMachine(const V3f &v) const
    {
        return V3f(Div.X*v.X, Div.Y*v.Y, Div.Z*v.Z);
    }

    bool verbose;
private:
    FT_HANDLE h;


};

#endif
