#pragma once
#include <math.h>

struct V3f;

struct V3i
{
    inline V3i() : X(0), Y(0), Z(0) {}
    inline V3i(const V3i &a)
    {
        X=a.X;
        Y=a.Y;
        Z=a.Z;
    }
    inline V3i(const V3f &f);

    inline V3i(int x, int y, int z) : X(x), Y(y), Z(z)
    {
    }

    int X, Y, Z;

    V3i operator+(const V3i &b) const
    {
        V3i c;
        c.X=X+b.X;
        c.Y=Y+b.Y;
        c.Z=Z+b.Z;
        return c;
    }
    V3i operator-(const V3i &b) const
    {
        V3i c;
        c.X=X-b.X;
        c.Y=Y-b.Y;
        c.Z=Z-b.Z;
        return c;
    }


    V3i &operator+=(const V3i &b)
    {
        X+=b.X;
        Y+=b.Y;
        Z+=b.Z;
        return *this;
    }
    V3i &operator-=(const V3i &b)
    {
        X-=b.X;
        Y-=b.Y;
        Z-=b.Z;
        return *this;
    }

    V3i &operator*=(int q)
    {
        X*=q;
        Y*=q;
        Z*=q;
        return *this;
    }

    V3i operator*(int q) const
    {
        V3i c;
        c.X=X*q;
        c.Y=Y*q;
        c.Z=Z*q;
        return c;
    }

    double Length() const
    {
        return sqrt((double)LengthSquare());
    }

    int LengthSquare() const
    {
        return X*X+Y*Y+Z*Z;
    }

    bool operator==(const V3i &b) const
    {
        return X==b.X && Y==b.Y && Z==b.Z;
    }
    bool operator!=(const V3i &b) const
    {
        return X!=b.X || Y!=b.Y || Z!=b.Z;
    }

};

inline V3i operator*(int q, V3i v)
{
    v.X*=q;
    v.Y*=q;
    v.Z*=q;
    return v;
}


struct V3f
{
    inline V3f() : X(0), Y(0), Z(0) {}
    inline V3f(const V3f &a)
    {
        X=a.X;
        Y=a.Y;
        Z=a.Z;
    }
    inline V3f(const V3i &a)
    {
        X=(double)a.X;
        Y=(double)a.Y;
        Z=(double)a.Z;
    }
    inline V3f(double x, double y, double z) : X(x), Y(y), Z(z)
    {
    }

    double X, Y, Z;

    V3f operator+(const V3f &b) const
    {
        V3f c;
        c.X=X+b.X;
        c.Y=Y+b.Y;
        c.Z=Z+b.Z;
        return c;
    }
    V3f operator-(const V3f &b) const
    {
        V3f c;
        c.X=X-b.X;
        c.Y=Y-b.Y;
        c.Z=Z-b.Z;
        return c;
    }

    V3f operator-() const
    {
        return V3f(-X, -Y, -Z);
    }

    V3f &operator+=(const V3f &b)
    {
        X+=b.X;
        Y+=b.Y;
        Z+=b.Z;
        return *this;
    }
    V3f &operator-=(const V3f &b)
    {
        X-=b.X;
        Y-=b.Y;
        Z-=b.Z;
        return *this;
    }

    V3f &operator*=(double q)
    {
        X*=q;
        Y*=q;
        Z*=q;
        return *this;
    }

    V3f operator*(double q) const
    {
        V3f c;
        c.X=X*q;
        c.Y=Y*q;
        c.Z=Z*q;
        return c;
    }

    double operator*(const V3f &b) const
    {
        return X*b.X+Y*b.Y+Z*b.Z;
    }

    V3f &Normalize()
    {
        return (*this)*=1/Length();
    }

    V3f Normalized() const
    {
        return (*this)*(1/Length());
    }

    V3f Tangent(V3f direction)
    {
        V3f N=direction.Normalized();
        return N*(N**this);
    }
    V3f Normal(V3f direction)
    {
        V3f N=direction.Normalized();
        return *this-N*(N**this);
    }

    double Length() const
    {
        return (double)sqrt(LengthSquare());
    }

    double LengthSquare() const
    {
        return X*X+Y*Y+Z*Z;
    }

    bool operator==(const V3f &b) const
    {
        return X==b.X && Y==b.Y && Z==b.Z;
    }
    bool operator!=(const V3f &b) const
    {
        return X!=b.X || Y!=b.Y || Z!=b.Z;
    }
    bool operator==(const V3i &b) const
    {
        return X==b.X && Y==b.Y && Z==b.Z;
    }
    bool operator!=(const V3i &b) const
    {
        return X!=b.X || Y!=b.Y || Z!=b.Z;
    }

};

inline V3i::V3i(const V3f &a)
{
    X=(int)a.X;
    Y=(int)a.Y;
    Z=(int)a.Z;
}

inline V3f operator*(double q, V3f v)
{
    v.X*=q;
    v.Y*=q;
    v.Z*=q;
    return v;
}

inline V3f operator*(double q, const V3i &v)
{
    V3f a;
    a.X=v.X*q;
    a.Y=v.Y*q;
    a.Z=v.Z*q;
    return v;
}

inline V3f CrossProduct(const V3f &a, const V3f &b)
{
    return V3f(
        a.Y*b.Z-b.Y*a.Z,
        a.Z*b.X-b.Z*a.X,
        a.X*b.Y-b.X*a.Y);
}

inline V3f Lerp(const V3f &A, const V3f &B, double t)
{
    return A+(B-A)*t;
}

// EOF
